//
// LDASTools frameCPP - A library implementing the LIGO/Virgo frame
// specification
//
// Copyright (C) 2022 California Institute of Technology
//
// LDASTools frameCPP is free software; you may redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 (GPLv2) of the
// License or at your discretion, any later version.
//
// LDASTools frameCPP is distributed in the hope that it will be useful, but
// without any warranty or even the implied warranty of merchantability
// or fitness for a particular purpose. See the GNU General Public
// License (GPLv2) for more details.
//
// Neither the names of the California Institute of Technology (Caltech),
// The Massachusetts Institute of Technology (M.I.T), The Laser
// Interferometer Gravitational-Wave Observatory (LIGO), nor the names
// of its contributors may be used to endorse or promote products derived
// from this software without specific prior written permission.
//
// You should have received a copy of the licensing terms for this
// software included in the file LICENSE located in the top-level
// directory of this package. If you did not, you can view a copy at
// http://dcc.ligo.org/M1500244/LICENSE
//

#define BOOST_TEST_MAIN
#include <boost/test/included/unit_test.hpp>

#include <cstdint>

#include <boost/range/irange.hpp>

#include "framecpp/Common/cksum.hh"
#include "framecpp/Common/CheckSum.hh"

bool init_ramped_buffer( );

using FrameCPP::Common::CheckSumCRC;
using FrameCPP::Common::cksum;

static const cksum::value_type expected_value_zero_0( 4294967295 );
static const cksum::value_type expected_value_zero_4096( 3018728591 );
static const cksum::value_type expected_value_zero_8589934592( 4135437457 );
static const cksum::value_type expected_value_ramped_4096( 300014538 );

uint8_t ramped_buffer[ 256 ];
bool    ramped_buffer_initted = init_ramped_buffer( );

bool
init_ramped_buffer( )
{
    for ( auto pos : boost::irange( size_t( 0 ), sizeof( ramped_buffer ) ) )
    {
        ramped_buffer[ pos ] = pos;
    }
    return ( true );
}

BOOST_AUTO_TEST_CASE( test_checksum_constructor )
{
    {
        static const auto expected_value( expected_value_ramped_4096 );
        CheckSumCRC       test_checksum( expected_value );

        BOOST_CHECK( test_checksum.value( ) == expected_value );
    }
}

BOOST_AUTO_TEST_CASE( test_checksum_zero_bytes )
{
    static const auto expected_value( expected_value_zero_0 );

    cksum       test_cksum;
    CheckSumCRC test_checksum;

    BOOST_CHECK( test_cksum.checksum( ) == expected_value );
    BOOST_CHECK( test_checksum.value( ) == expected_value );
}

BOOST_AUTO_TEST_CASE( test_checksum_4096_bytes )
{
    static const auto expected_value( expected_value_zero_4096 );

    cksum       test_cksum;
    CheckSumCRC test_checksum;

    uint8_t zero_buffer[ 1024 ];
    std::fill( zero_buffer, &( zero_buffer[ sizeof( zero_buffer ) ] ), 0 );

    for ( auto i : boost::irange( 0, 4 ) )
    {
        test_cksum.process_bytes( zero_buffer, sizeof( zero_buffer ) );
        test_checksum.calc( zero_buffer, sizeof( zero_buffer ) );
    }

    BOOST_CHECK( test_cksum.checksum( ) == expected_value );
    BOOST_CHECK( test_checksum.value( ) == expected_value );
}

BOOST_AUTO_TEST_CASE( test_checksum_4096_bytes_2_stages )
{
    static const auto expected_value( expected_value_zero_4096 );

    uint8_t zero_buffer[ 1024 ];
    std::fill( zero_buffer, &( zero_buffer[ sizeof( zero_buffer ) ] ), 0 );

    cksum       test_cksum_first;
    CheckSumCRC test_checksum_first;

    for ( auto i : boost::irange( 0, 2 ) )
    {
        test_cksum_first.process_bytes( zero_buffer, sizeof( zero_buffer ) );
        test_checksum_first.calc( zero_buffer, sizeof( zero_buffer ) );
    }

    cksum        test_cksum{ test_cksum_first };
    CheckSumCRC* test_checksum{ test_checksum_first.Clone( ) };

    for ( auto i : boost::irange( 2, 4 ) )
    {
        test_cksum.process_bytes( zero_buffer, sizeof( zero_buffer ) );
        test_checksum->calc( zero_buffer, sizeof( zero_buffer ) );
    }

    BOOST_CHECK( test_cksum.checksum( ) == expected_value );
    BOOST_CHECK( test_checksum->value( ) == expected_value );
}

BOOST_AUTO_TEST_CASE( test_checksum_4096_bytes_reset )
{
    static const auto expected_value( expected_value_zero_4096 );

    cksum       test_cksum;
    CheckSumCRC test_checksum;

    uint8_t zero_buffer[ 1024 ];
    std::fill( zero_buffer, &( zero_buffer[ sizeof( zero_buffer ) ] ), 0 );

    for ( auto i : boost::irange( 0, 8 ) )
    {
        test_cksum.process_bytes( zero_buffer, sizeof( zero_buffer ) );
        test_checksum.calc( zero_buffer, sizeof( zero_buffer ) );
    }

    test_cksum.reset( );
    test_checksum.Reset( );

    for ( auto i : boost::irange( 0, 4 ) )
    {
        test_cksum.process_bytes( zero_buffer, sizeof( zero_buffer ) );
        test_checksum.calc( zero_buffer, sizeof( zero_buffer ) );
    }

    BOOST_CHECK( test_cksum.checksum( ) == expected_value );
    BOOST_CHECK( test_checksum.value( ) == expected_value );

    test_cksum.reset( );
    test_checksum.Reset( );

    std::ofstream ramped_4096( "ramped_4096.dat", std::ofstream::out );
    for ( auto buffer_count : boost::irange(
              size_t( 0 ), size_t( 4096 / sizeof( ramped_buffer ) ) ) )
    {
        ramped_4096.write( reinterpret_cast< const char* >( ramped_buffer ),
                           sizeof( ramped_buffer ) );
        test_cksum.process_bytes( ramped_buffer, sizeof( ramped_buffer ) );
        test_checksum.calc( ramped_buffer, sizeof( ramped_buffer ) );
    }
    ramped_4096.close( );

    BOOST_CHECK( test_cksum.checksum( ) == expected_value_ramped_4096 );
    BOOST_CHECK( test_checksum.value( ) == expected_value_ramped_4096 );
}

BOOST_AUTO_TEST_CASE( test_checksum_4096_bytes_no_extend )
{
    static const auto expected_value( expected_value_zero_4096 );

    cksum       test_cksum;
    CheckSumCRC test_checksum;

    uint8_t zero_buffer[ 1024 ];
    std::fill( zero_buffer, &( zero_buffer[ sizeof( zero_buffer ) ] ), 0 );

    for ( auto i : boost::irange( 0, 4 ) )
    {
        test_cksum.process_bytes( zero_buffer, sizeof( zero_buffer ) );
        test_checksum.calc( zero_buffer, sizeof( zero_buffer ) );
    }

    BOOST_CHECK( test_cksum.checksum( ) == expected_value );
    BOOST_CHECK( test_checksum.value( ) == expected_value );

    for ( auto i : boost::irange( 0, 4 ) )
    {
        test_cksum.process_bytes( zero_buffer, sizeof( zero_buffer ) );
        test_checksum.calc( zero_buffer, sizeof( zero_buffer ) );
    }

    BOOST_CHECK( test_cksum.checksum( ) == expected_value );
    BOOST_CHECK( test_checksum.value( ) == expected_value );
}

BOOST_AUTO_TEST_CASE( test_checksum_8589934592_bytes )
{
    // Test that can calculate checksums for streams greater
    // thsn 2^32 bytes in length.

    static const auto expected_value( expected_value_zero_8589934592 );

    cksum       test_cksum;
    CheckSumCRC test_checksum;

    uint8_t zero_buffer[ 65536 ];
    std::fill( zero_buffer, &( zero_buffer[ sizeof( zero_buffer ) ] ), 0 );

    for ( auto i : boost::irange(
              size_t( 0 ), size_t( 8589934592L / sizeof( zero_buffer ) ) ) )
    {
        test_cksum.process_bytes( zero_buffer, sizeof( zero_buffer ) );
        test_checksum.calc( zero_buffer, sizeof( zero_buffer ) );
    }

    BOOST_CHECK( test_cksum.checksum( ) == expected_value );
    BOOST_CHECK( test_checksum.value( ) == expected_value );
}
